; Version 0.1

#include "usb-devices.inc"
#include "jdlib.inc"

	#define	USBPARITY	(0x08 & 0xFF)
	#define USE_CTSDTR	1

; Optimized CASE statement, where the next block is a DEFAULT with a single
; branch instruction.  This complements the macros in jdlib.inc
CASEELSE macro n, addr
#if (g_swaddr#v(g_swdpth)!=g_swnext#v(g_swdpth))
	bra	swend#v(g_swaddr#v(g_swdpth))
#endif
swnxt#v(g_swnext#v(g_swdpth))
g_swnext#v(g_swdpth)	set g_swnext#v(g_swdpth)+1
#if (n^g_switch#v(g_swdpth) != 0)
	xorlw	(n ^ g_switch#v(g_swdpth))
g_switch#v(g_swdpth)	set n
#endif
	bnz		addr
	endm

; Offsets into standard USB descriptors
bmRequestType	equ	0x00
bRequest		equ	0x01
wValueL			equ	0x02
wValueH			equ	0x03
wIndexL			equ	0x04
wIndexH			equ	0x05
wLengthL		equ	0x06
wLengthH		equ	0x07

; Standard USB descriptor types
DESC_DEVICE			equ 0x01
DESC_CONFIG			equ 0x02
DESC_STRING			equ 0x03
DESC_INTERFACE		equ 0x04
DESC_ENDPOINT		equ	0x05

; USB Entpoint types
ENDPT_CTRL			equ 0x00
ENDPT_ISO			equ 0x01
ENDPT_BULK			equ 0x02
ENDPT_INT			equ 0x03

CDC_CLASS			equ	0x02
CDC_COMMINTF		equ	0x02
CDC_ABSTRACTCONTROL	equ	0x02
CDC_V25TER			equ	0x01
CDC_DATAINTFCLASS	equ	0x0A
CDC_NOPROTOCOL		equ	0x00
CDC_CS_INTERFACE	equ	0x24
;CDC_CS_ENDPOINT		equ	0x25
CDC_DSC_FNHEAD		equ	0x00
CDC_DSC_FNCALLMGMT	equ	0x01
CDC_DSC_FNACM		equ	0x02
CDC_DSC_FNUNION		equ	0x06

; USB Request types, also stored in USB_request
USB_SETADDR	equ	0xF5
USB_GETDESC	equ	0xF6
USB_SETCONF equ 0xF9
USB_SETINTR equ 0xFB

;CDC Class Requests
;CDC_SENDCMD			equ 0x00
;CDC_GETRESP			equ	0x01
;CDC_SETFEAT			equ 0x02
;CDC_GETFEAT			equ	0x03
;CDC_CLEARFEAT		equ 0x04
CDC_SETLINECODING	equ	0x20
CDC_GETLINECODING	equ	0x21
CDC_SETCONTROLSTA	equ	0x22
;CDC_SENDBREAK		equ 0x23

; Defines for accessing USB Buffer descriptors (as per PIC18F datasheet)
; DESC____ is the offset for each of the 4 registers
DESCSTAT	equ	0x00	;BDnSTAT in PIC18F spec
DESCBCNT	equ	0x01	;BDnCNT in PIC18F spec
DESCADRL	equ	0x02	;BDnADRL in PIC18F spec
DESCADRH	equ	0x03	;BDnADRH in PIC18F spec

; Address of buffer descriptor for Ping/Pong buffering mode
; UBD0Io - 0 is endpoint number, I is direction (I)n or (O)ut
; o indicates Odd PingPong buffer, Even is unmarked
UBD0O		equ	UBDESC+0x00	;Endpoint 0, Out, Even
UBD0Oo		equ	UBDESC+0x04	;Endpoint 0, Out, Odd
UBD0I		equ	UBDESC+0x08	;Endpoint 0, In, Even
UBD0Io		equ	UBDESC+0x0c	;Endpoint 0, In, Odd
UBD1O		equ	UBDESC+0x10	;Endpoint 1
UBD1Oo		equ	UBDESC+0x14
UBD1I		equ	UBDESC+0x18
UBD1Io		equ	UBDESC+0x1c
UBD2O		equ	UBDESC+0x20	;Endpoint 2
UBD2Oo		equ	UBDESC+0x24
UBD2I		equ	UBDESC+0x28
UBD2Io		equ	UBDESC+0x2c

; Per-buffer sizes for each endpoint.
EP0SIZE		equ	0x10
CDCINTSIZE	equ	0x08	; EP1 - CDC rts/cts code requires this stay fixed
CDCBULKSIZE	equ	0x20	; EP2 - CDC send code requires this stay fixed

; USB buffer addresses.  4 buffers for EP0, 2 for EP1, and 4 for EP2
EP0BUFF	equ	0x30		; 0x30 - provides room for endpoints 0,1 and 2
EP1BUFF equ 0x70
EP2BUFF equ 0x80		; CDC send code requires this stay fixed

; RAM allocations.  Places ram at the end of memory defined by RAMTOP in usb-devices.inc
	cblock	RAMTOP - 0x0C; use last 12 bytes of ram
	USB_request			; current USB request being serviced
	USB_request_val		; data for associated usb request (usually byte count)
	USB_state			; usb state, high bit = ping pong even/odd
	tempreg0			; temporary registers (only used by baud divider)
	tempreg1
	tempreg2
	tempreg3
	baudrate0			; stores current baud rate
	baudrate1
	baudrate2
	TX_cntr				; counts bytes to be transmitted
	ctsdtr				; last state of input CTS/DTR signals
	endc


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Program code starts here
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Note: The reset/low_int/high_int vectors are jumbled together somewhat to
; pack the code in the start of flash without gaps.
; Apologies if this is a little harder to follow
	org 0x000000				; RESET_VECTOR: Execution begins here on reset
	bsf		RCON, IPEN, ACCESS	; Enable Interrupt Priority
	movlw	0x20
	movwf	IPR1, ACCESS		; All priorities to low (except RCIF)
	bra		boot_main			; Continue boot initialization


	org 0x000008					;High Priority interrupt handler
	movff	RCREG, POSTINC0
	
	movf	FSR0L, W, ACCESS	; check if buffer full
	andlw	0x1F				; Each buffer is 0x20 in length
	bnz		hint_done			; terminate if buffer is not full
	
	banksel	UBDESC				; otherwise, a buffer is full, so send it
	movlw	CDCBULKSIZE			; Buffer size
	bra		hint				; Continue High Priority interrupt handler


	org 0x000018					;Low Priority interrupt handler
	btfsc	UIR, SOFIF, ACCESS		;Start of frame token recvd
	bra	int_sof
	btfsc	UIR, IDLEIF, ACCESS		;Idle condition detected
	bra	int_idle
	btfsc	UIR, ACTVIF, ACCESS		;Activity on data lines detected
	bra	int_actv
;	btfsc	UIR, STALLIF, ACCESS	;STALL handshake sent by SIE
;	bra	int_stall
	btfsc	UIR, URSTIF, ACCESS		;USB reset occured
	reset
	btfsc	UIR, TRNIF, ACCESS		;USB Transaction Complete
	bra	int_xfer

	org	0x002C						; High Priority interrupt handler (continued)
hint:
	btfss	FSR0L, 5, ACCESS	; determine if even or odd buffer is filled
	bra		hint_odd
	movwf	UBD2I+DESCBCNT, BANKED		; set packet size
	movlw	0x80; | USBPARITY
	movwf	UBD2I+DESCSTAT, BANKED		; and hand over to SIE
	bra		hint_done
hint_odd:
	movwf	UBD2Io+DESCBCNT, BANKED		; set packet size
	movlw	0xC0; | USBPARITY
	movwf	UBD2Io+DESCSTAT, BANKED		; and hand over to SIE
	lfsr	FSR0, 0x4C0				; Reset input buffer to start 
hint_done:
	retfie	FAST

device_desc
	db	0x12, DESC_DEVICE	; bLength, bDescriptorType
	db	0x00, 0x02			; bcdUSB (low byte), bcdUSB (high byte)
	db	CDC_CLASS, 0x00		; bDeviceClass, bDeviceSubClass
	db	0x00, EP0SIZE		; bDeviceProtocl, bMaxPacketSize
	db	0xD8, 0x04			; idVendor (low byte), idVendor (high byte)
	db	0x0a, 0x00			; idProduct (low byte), idProduct (high byte)
	db	0x01, 0x00			; bcdDevice (low byte), bcdDevice (high byte)
	db	0x00, 0x00			; iManufacturer, iProduct
	db	0x00, 0x01			; iSerialNumber, bNumConfigurations
	
config_desc
	db	0x09, DESC_CONFIG					;02	bLength, bDescriptorType
	db	0x43, 0x00							;04 wTotalLength low, high
	db	0x02, 0x01							;06	bNumInterfaces, bConfigurationValue
	db	0x00, 0x80							;08 iConfiguration, bmAttributes
	db	0x32, 0x09							;0A bMaxPower, bLength
	
	db	DESC_INTERFACE, 0x00				;0C bDescriptorType, bInterfaceNumber
	db	0x00, 0x01							;0E bAlternateSetting, bNumEndpoints
	db	CDC_COMMINTF, CDC_ABSTRACTCONTROL	;10 bInterfaceClass, bInterfaceSubclass
	db	CDC_V25TER, 0x00					;12 bInterfaceProtocol, iInterface

	db	0x05, CDC_CS_INTERFACE				;14 bLength, bDescriptorType
	db	CDC_DSC_FNHEAD, 0x10				;16 bDescriptorSubType, bcdCDC low
	db	0x01, 0x05							;18 high, bLength

	db	CDC_CS_INTERFACE, CDC_DSC_FNCALLMGMT;1A bDescriptorType, bDescriptorSubType
	db	0x01, 0x01							;1C bmCapabilities, bDataInterface

	db	0x04, CDC_CS_INTERFACE				;1E bLength, bDescriptorType
	db	CDC_DSC_FNACM, 0x02					;20 bDescriptorSubType, bmCapabilities
	
	db	0x05, CDC_CS_INTERFACE				;22 bLength, bDescriptorType
	db	CDC_DSC_FNUNION, 0x00				;24 bDescriptorSubType, bMasterIntf
	db	0x01, 0x07							;26 bSlaveIntf, bLength

	db	DESC_ENDPOINT, 0x81					;28 bDescriptorType, bEndpointAddress
	db	ENDPT_INT, CDCINTSIZE				;2A bmAttributes, wMaxPacketSize low
	db	0x00, 0x0A							;2C high, bInterval
	
	db	0x09, DESC_INTERFACE				;2E bLength, bDescriptorType
	db	0x01, 0x00							;30 bInterfaceNumber, bAlternateSetting
	db	0x02, CDC_DATAINTFCLASS				;32 bNumEndpoints, bInterfaceClass
	db	0x00, CDC_NOPROTOCOL				;34 bInterfaceSubClass, bInterfaceProtocol
	db	0x00, 0x07							;36 iInterface string, bLength
	
	db	DESC_ENDPOINT, 0x02					;38 bDescriptorType, bEndpointAddress
	db	ENDPT_BULK, CDCBULKSIZE				;3A bmAttributes, wMaxPacketSize low
	db	0x00, 0x00							;3C high, bInterval
	
	db	0x07, DESC_ENDPOINT					;3E bLength, bDescriptorType
	db	0x82, ENDPT_BULK					;40 bEndpointAddress, bmAttributes
	db	CDCBULKSIZE, 0x00					;42 wMaxPacketSize low, high
	db	0x00								;43 bInterval (+1 filler byte)

if	USE_CTSDTR
serial_state			; Serial_state header used to report CTS/DSR changes
	db	0xA1, 0x20			; bmRequestType, bRequest
	db	0x00, 0x00			; wValueL, wValueH
	db	0x01, 0x00			; wIndexL, wIndexH
	db	0x02, 0x00			; wLengthL, wLengthH
endif

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Start of Frame interrupt handler - triggered every 1ms by USB host SOF message
; Sends a partial CDC data packet if neither buffer is being used by USB-SIE
; Also checks for changes to input CTS&DSR signals and sends serial_state packet
; on EP1 if a change is detected
; Worst case runtime = 
int_sof
	bcf		PIE1, RCIE, ACCESS			;disable serial recieve interrupt
	banksel	UBDESC
	movlw	0x1F
	andwf	FSR0L, W, ACCESS			;check if we have any pending bytes
	bz		sof_nodata
	btfsc	UBD2I+DESCSTAT, 7, BANKED	;check if SIE owns either input buffer
	bra		sof_nodata
	btfsc	UBD2Io+DESCSTAT, 7, BANKED	;check if SIE owns either input buffer
	bra		sof_nodata
	btfsc	FSR0L, 5, ACCESS			;determine if even or odd buffer is current
	bra		sof_odd
	movwf	UBD2I+DESCBCNT, BANKED		; set packet size
	movlw	0x80 | USBPARITY
	movwf	UBD2I+DESCSTAT, BANKED		; and hand over to SIE
	bra		sof_advance
sof_odd:
	movwf	UBD2Io+DESCBCNT, BANKED		; set packet size
	movlw	0xC0 | USBPARITY
	movwf	UBD2Io+DESCSTAT, BANKED		; and hand over to SIE
sof_advance:
	movlw	0xE0
	andwf	FSR0L, F, ACCESS			; move to start of buffer
	btg		FSR0L, 5, ACCESS			; and switch buffers

sof_nodata
	bsf	PIE1, RCIE, ACCESS				; Enable serial recieve interrupt

if	USE_CTSDTR
	;Do Ctsdtr
	banksel	USB_state
	rrncf	PORTB, W, ACCESS			; check if input CTS/DSR signals have changed
	xorwf	ctsdtr, W, BANKED
	andlw	0x06
	bz		sof_exit
	xorwf	ctsdtr, F, BANKED			; if they have, update cached copy
	lfsr	FSR1, UBD1I					; and send serial_state message on EP1
	rcall	fsr2_to_fsr1				; point FSR2 for EP1I's data buffer
	movlw	serial_state
	movwf	TBLPTRL, ACCESS
	movlw	0x08
	rcall	send_flashbytes				; load the 8-byte serial_state header
	rrncf	ctsdtr, W, BANKED
	andlw	0x03
	movwf	POSTINC2, ACCESS			;UART state
	clrf	POSTINC2, ACCESS
	
	btg		FSR1L, 2, ACCESS			; advance to odd BDESC (UBD1Io)
	movlw	0x02						; respond with 10 bytes (8 above, 2 here)
	rcall	handover_data1

sof_exit
endif
	bcf	UIR, SOFIF, ACCESS				; clear SOF interrupt bit
	bra int_res_exi						; and exit handler


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; USB Idle interrupt
int_idle
	bcf	UIR, IDLEIF, ACCESS
	bsf	UCON, SUSPND, ACCESS	;Device is idle, so suspend usb
	bra	int_res_exi
	
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; USB Activity interrupt (wake from idle)
int_actv
	bcf	UCON, SUSPND, ACCESS	;Activity detected, so wake up usb
actv_l
	bcf	UIR, ACTVIF, ACCESS
	btfsc	UIR, ACTVIF, ACCESS
	bra	actv_l					;Loop until ACTVIF is clear
	bra	int_res_exi

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; USB Xfer completed interrupt
; All USB transactions completed (Setup/In/Out) will trigger this interrupt
; Determine what transaction completed, and take the apropriate action
int_xfer
	rcall	fsr12_to_ustat		; FSR1/FSR2 point to UBDESC & Buffer for xfer

	banksel USB_request

	movf	INDF1, W, ACCESS	; Load BDxSTATUS into W
	andlw	0x3c				; Extract PID
	SWITCHW
	CASE	0x34				; PID=Setup packet
		banksel	UBDESC
		clrf	UBD0I+DESCSTAT, BANKED		; Dequeue any pending in packets
		clrf	UBD0Io+DESCSTAT, BANKED
		
		banksel	USB_request

        movlw	EP0SIZE			; Hand buffer back to USB-SIE
		rcall	handover_data0

		movf	[bmRequestType], W
		andlw	0x80			; Check if request is IN or OUT bound
		bnz	seteve

		movf	[wLengthL], W	; Check if data length is zero
		bnz	setodd
		movf	[wLengthH], W	; Check if data length is zero
		bnz	setodd
		bra	seteve

setodd							; Set data1 bit on ODD endpoint buffer descriptor
		movlw	0xC0 | USBPARITY
		btg	FSR1L, 2, ACCESS
		movwf	INDF1, ACCESS
		btg	FSR1L, 2, ACCESS
seteve
		movf	[bmRequestType], W
		andlw	0x60		; 00 = Standard, 01=Class, 10 = Vendor, 11=error
		SWITCHW
		CASE	0x00						; bmRequestType = Standard
			movf	[bRequest], W			; Get bRequest from EP
			SWITCHW
			CASE	(USB_SETADDR & 0x0f)	; is bRequest=SET_ADDRESS
				movf	[wValueL], W		; Get address from EP wValue
				movwf	USB_request_val, BANKED
				movlw	USB_SETADDR
				movwf	USB_request, BANKED	; USB_request = Set Address

			CASE	(USB_GETDESC & 0x0f)	; is bRequest=GET_DESCRIPTOR
				movf	[wValueH], W		; get descriptor type
				SWITCHW
				CASE	DESC_DEVICE
					movlw	device_desc		; Device descriptor requested
					movwf	TBLPTRL, ACCESS
					movlw	0x12			; desc length
				CASEELSE	DESC_CONFIG, setup_error	; setup_error if not desc/conf
					movlw	config_desc		; Config descriptor requested
					movwf	TBLPTRL, ACCESS
					movlw	0x43			; desc length
				ENDSW
				movwf	USB_request_val, BANKED
				movlw	0xFF				; Windows driver requests 0x109 bytes
				btfss	[wLengthH],0		; but only sending 0xFF is supported
				movf	[wLengthL], W		; Find the minimum, of bytes reqd & desc len
				subwf	USB_request_val, W, BANKED
				bnc		$+4
				subwf	USB_request_val, F, BANKED
				movlw	USB_GETDESC
				movwf	USB_request, BANKED	; USB_request = Get Descriptor

				rcall	fsr12_to_ep0in		; Point FSR1/FSR2 to EP0 In

				rcall	send_descriptor		; And send descriptor
				bra		setup_done

			CASE	(USB_SETINTR & 0x0f)	; is bRequest=SET_INTERFACE
											; do nothing

			CASEELSE	(USB_SETCONF & 0x0F), setup_error; is bRequest = SET_CONFIG
				clrf	UEP1, ACCESS		;Reset Endpoints 1 & 2
				clrf	UEP2, ACCESS

				movlw	0x1E				;Activate Endpoints 1 & 2
				movwf	UEP1, ACCESS
				movwf	UEP2, ACCESS

				bsf		PIE1, RCIE, ACCESS	;Enable serial Recieve interrupt
			ENDSW
setdone_xmitzero
			rcall	fsr12_to_ep0in			; Point FSR1/FSR2 to EP0 In

			movlw	0x00					; and send zero bytes
			rcall	handover_data1
		CASE	0x20						; bmRequestType = Class
			movf	[bRequest], W			; Get bRequest from EP
			SWITCHW
			CASE	CDC_SETLINECODING		; bRequest = CDC Set Line Coding
				movlw	CDC_SETLINECODING
				movwf	USB_request, BANKED	; USB_request = CDC_SETLINECODING
if	USE_CTSDTR
			CASE	CDC_GETLINECODING		; bRequest = CDC Get Line Coding
else
			CASEELSE	CDC_GETLINECODING, setup_error
endif
				movlw	CDC_GETLINECODING
				movwf	USB_request, BANKED	; USB_Request = CDC_GETLINECODING
				
				rcall	fsr12_to_ep0in		; Point FSR1/FSR2 to EP0 In
				movff	baudrate0, POSTINC2	; Send baud rate
				movff	baudrate1, POSTINC2
				movff	baudrate2, POSTINC2
				
				clrf	POSTINC2			;Baud 4th byte
				clrf	POSTINC2			;Stop Bits 0=1
				clrf	POSTINC2			;Parity Bits 0=None
				movlw	0x08
				movwf	POSTINC2			;8 bits per byte

				movlw	0x07				;send 7 bytes
				rcall	handover_data1

if	USE_CTSDTR
			CASEELSE	CDC_SETCONTROLSTA, setup_error ; bRequest = CDC_SETCONTROLSTA
				movf	[wValueL], W		;Get DTR/RTS from request
				xorwf	LATB, W, ACCESS		;And update LATB with result
				andlw	0x03
				xorwf	LATB, F, ACCESS		;bit 0=DTR, 1=RTS
				bra	setdone_xmitzero
endif
			ENDSW
		DEFAULT								; All other cases should trigger stall
setup_error									;Error occurred, stall EP0
			movlw	EP0SIZE
			movwf	PREINC1, ACCESS
			decf	FSR1L, F, ACCESS
			movlw	0x84 | USBPARITY
			movwf	INDF1, ACCESS			; set output buffer stall

			lfsr	1, UBD0I+DESCSTAT
			btfsc	USB_state, 7, BANKED
			btg	FSR1L, 2, ACCESS
			movlw	0x84 | USBPARITY
			movwf	INDF1, ACCESS			; set input buffer stall
		ENDSW
setup_done
		bcf	UCON, PKTDIS, ACCESS			;enable USB SIE token & packet processing

	CASE	0x24							; PID=In packet
		movf	USB_request, W, BANKED
		SWITCHW
		CASE	USB_SETADDR					; USB_request = USB_SETADDR
			clrf	USB_request, BANKED		; clear USB_request
			movf	USB_request_val, W, BANKED
			movwf	UADDR, ACCESS			; And set USB Address

		CASE	USB_GETDESC					; USB_request = USB_GETDESC
			movf	USB_request_val, W, BANKED
			bz	in_upload_done				; do nothing if all data sent
			rcall	fsr12_to_ep0in			; Point FSR1/FSR2 to EP0 In
			rcall	send_descriptor			; otherwise send next descriptor packet

		CASE	CDC_GETLINECODING			; USB_request = CDC_GETLINECODING
in_upload_done
			clrf	USB_request, BANKED		; no data left to transfer clear USB_request
		ENDSW

	CASE	0x04							; PID=Out packet
		movf	USTAT, W, ACCESS			; USTAT= which endpoint/buffer triggered int
		andlw	0x78						;Get endpoint
		SWITCHW
		CASE	0x10						;Endpoint = EP2 (CDC data)
			movf	PREINC1, W, ACCESS		;Get number of bytes recvd
			movwf	TX_cntr, BANKED
			bz		ser_skip				;If no bytes recieved, skip(is this needed?)
ser_loop
			movf	POSTINC2, W, ACCESS		;Get next byte recvd to WREG
ser_wait
			btfss	TXSTA, 0x1, ACCESS		;Wait for TXREG empty (last byte completed)
			bra		ser_wait
			movwf	TXREG, ACCESS			;put byte recvd to TXREG (serial transmit)
			decfsz	TX_cntr, F, BANKED		;decrement number of bytes to send
			bra		ser_loop				; loop over remaining bytes
ser_skip
			movlw	CDCBULKSIZE				;reset EP2O buffer size
			movwf	INDF1, ACCESS
			movlw	0x80 | USBPARITY
			btfsc	FSR1L, 2, ACCESS		;test if even/odd buffer
			iorlw	0x40					;set data1 bit if odd buffer
			decf	FSR1L, F, ACCESS
			movwf	INDF1, ACCESS			;return EP2O buffer to USB SIE

		DEFAULT								;Endpoint != EP2 (must be EP0)
			movf	USB_request, W, BANKED
			SWITCHW
			CASE	CDC_SETLINECODING		;USB_request = CDC_SETLINECODING
				clrf	USB_request, BANKED	;clear USB_request (done after this)
				bcf		RCSTA, SPEN, ACCESS	;disable serial port
				movff	POSTINC2, baudrate0	;copy new baudrate
				movff	POSTINC2, baudrate1
				movff	POSTINC2, baudrate2
				rcall	DoBaudDiv			;set new baudrate on serial
				bsf		RCSTA, SPEN, ACCESS	;enable serial port
			ENDSW
			
			movlw	EP0SIZE					; Return buffer to SIE
			rcall	handover_data0

			bra		setdone_xmitzero		; Send zero byte acknowledgement
		ENDSW
	ENDSW

xferi_exit
    bcf		UIR, TRNIF, ACCESS				; Clears USTAT
int_res_exi
	bcf		PIR2, USBIF, ACCESS				; clear usb interrupt flag
	retfie	0								;Exit interrupt handler
	

;;;;;;;;;;;; end transmit interrupt code

boot_main:
	clrf	IPR2, ACCESS			; interrupt priorities to low
	
	movlw	0x16			;Pullup Enable, Full Speed, PingPong on all endpoints
	movwf	UCFG, ACCESS

	bsf		UCON, USBEN, ACCESS		;Enable USB SIE
	
	lfsr	FSR1, UEP0				;FSR1 to UEP0, to clear all 16 in loop
	movlw	0x10					;set loop counter
init_clr_uep
	bcf		UIR, TRNIF, ACCESS		;Clear USTAT queue
	clrf	POSTINC1, ACCESS		;Clear UEP0 through UEP15
	btg		UCON, PPBRST, ACCESS	;Reset ping pong buffers
	decfsz	WREG, F, ACCESS			;decrement loop counter
	bra		init_clr_uep			;and loop

	banksel	UBDESC
	movlw	0x04					; Buffer high address
	movwf	UBD0O+DESCADRH, BANKED
	movwf	UBD0Oo+DESCADRH, BANKED
	movwf	UBD0I+DESCADRH, BANKED
	movwf	UBD0Io+DESCADRH, BANKED
	movwf	UBD1I+DESCADRH, BANKED;EP1_i
	movwf	UBD1Io+DESCADRH, BANKED
	movwf	UBD2O+DESCADRH, BANKED;EP2_o
	movwf	UBD2Oo+DESCADRH, BANKED
	movwf	UBD2I+DESCADRH, BANKED;EP2_i
	movwf	UBD2Io+DESCADRH, BANKED

	movlw	EP0BUFF					; Buffer low address
	movwf	UBD0O+DESCADRL, BANKED
	movlw	EP0BUFF + EP0SIZE
	movwf	UBD0Oo+DESCADRL, BANKED
	movlw	EP0BUFF + (EP0SIZE * 2)
	movwf	UBD0I+DESCADRL, BANKED
	movlw	EP0BUFF + (EP0SIZE * 3)
	movwf	UBD0Io+DESCADRL, BANKED
	movlw	EP1BUFF
	movwf	UBD1I+DESCADRL, BANKED
	movlw	EP1BUFF + CDCINTSIZE
	movwf	UBD1Io+DESCADRL, BANKED
	movlw	EP2BUFF
	movwf	UBD2O+DESCADRL, BANKED
	movlw	EP2BUFF + CDCBULKSIZE
	movwf	UBD2Oo+DESCADRL, BANKED
	movlw	EP2BUFF + (CDCBULKSIZE * 2)
	movwf	UBD2I+DESCADRL, BANKED
	movlw	EP2BUFF + (CDCBULKSIZE * 3)
	movwf	UBD2Io+DESCADRL, BANKED
	
	movlw	EP0SIZE					; Buffer sizes (for out buffers only)
	movwf	UBD0O+DESCBCNT, BANKED
	movwf	UBD0Oo+DESCBCNT, BANKED
	movlw	CDCBULKSIZE
	movwf	UBD2O+DESCBCNT, BANKED
	movwf	UBD2Oo+DESCBCNT, BANKED
	
	movlw	0x80 | USBPARITY		; Pass buffer ownership to USB-SIE
	movwf	UBD0O+DESCSTAT, BANKED	;  (out buffers only)
	movwf	UBD2O+DESCSTAT, BANKED
	movlw	0xC0 | USBPARITY
	movwf	UBD0Oo+DESCSTAT, BANKED
	movwf	UBD2Oo+DESCSTAT, BANKED

	movlw	0x40 | USBPARITY	; Set data1 on EP1,In,Odd (needed for send_flashbytes)
	movwf	UBD1Io+DESCSTAT, BANKED
	
	clrf	UBD0I+DESCSTAT, BANKED	; Safety measure, claim ownership of all input EPt
	clrf	UBD0Io+DESCSTAT, BANKED
	clrf	UBD1I+DESCSTAT, BANKED
	clrf	UBD2I+DESCSTAT, BANKED
	clrf	UBD2Io+DESCSTAT, BANKED

	banksel	USB_request
	clrf	USB_request, BANKED		; Clear USB_request and USB_state
	clrf	USB_state, BANKED

;	setf	UEIE, ACCESS			; Enable all USB Error interrupts
	bcf		PIR2, USBIF, ACCESS		; Clear USB interrupt flag
	movlw	0x16
	movwf	UEP0, ACCESS			; Enable USB endpoint 0

se0wait
	btfsc	UCON, SE0, ACCESS		; Wait for USB single ended zero to end
	bra	se0wait
	
	movlw	0x5d
	movwf	UIE, ACCESS				;Enable All USB Interrupts except uerri & stallie
	bsf		PIE2, USBIE, ACCESS

baud EQU .115200					; the default baud rate
	movlw	upper baud				;Initialize baud rate (not actually required)
	movwf	baudrate2
	movlw	high baud
	movwf	baudrate1
	clrf	baudrate0
	rcall	DoBaudDiv
	
	bsf		BAUDCON, BRG16, ACCESS	;Use 16-bit baudrate divider
	movlw 	b'00100100'				;TXEN | BRGH
	movwf 	TXSTA, ACCESS
	movlw	b'10010000'				;SPEN | CREN
	movwf	RCSTA, ACCESS

	lfsr	FSR0, 0x4C0				;Point FSR0 at EP2 In buffer (for serial Recv)

	movlw	b'11111100'
	movwf	TRISB, ACCESS			;Use PORTB for RTS/CTS
	clrf	LATB, ACCESS			;and set both to off

if	USE_CTSDTR
	clrf	ctsdtr, BANKED			;Clear 'previous' CTS/DSR state
endif
	
	movlw	0xc0
	movwf	INTCON, ACCESS			;Enable High&Low Priority Ints

	bra		$						;And loop here forever

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; send descriptor sends what's pointed to by the tblptrs, up to USB_request_val
; bytes, no more than ep0 size
send_descriptor
	movlw	EP0SIZE					;Check if requested > buffer
	subwf	USB_request_val, W, BANKED
	bnn	desc_too_big
	movf	USB_request_val, W, BANKED
	clrf	USB_request_val, BANKED	;Clear requested if we finish this round
	bra	send_flashbytes
desc_too_big
	movwf	USB_request_val, BANKED	;Decrement requested by amount sent
	movlw	EP0SIZE					;and send buffer_size bytes
	; note this continues into the next function

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; send flashbytes sends what's pointed to by the tblptrs, and sends WREG bytes
; does not check against buffer size
send_flashbytes
	movwf	PREINC1, ACCESS			;Put number of bytes into DESCBCNT
	decf	FSR1L, F, ACCESS
sdloop	tblrd	*+					;read byte from program memory
	movff	TABLAT, POSTINC2		;place byte into FSR2 buffer
	addlw	0xff					; decrement W
	bnz	sdloop						; and loop through rememaining bytes

	movlw	0x40
	btg	FSR1L, 2, ACCESS
	xorwf	INDF1, W, ACCESS		;determine correct data0/data1 bit
	btg	FSR1L, 2, ACCESS
	andlw	0x40
	iorlw	0x80 | USBPARITY
	movwf	INDF1, ACCESS			;hand over INPBUFF to SIO

	return	

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; fsr12_to_ep0in points FSR1 at UBDESC for EP0 in, and then FSR2 at its buffer
fsr12_to_ep0in
	lfsr	FSR1, UBD0I+DESCSTAT
	btfsc	USB_state, 7, BANKED
	btg	FSR1L, 2, ACCESS		;switch BuffDesc (even/odd) based on ping/pong
	btg	USB_state, 7, BANKED
	bra	fsr2_to_fsr1			; continues in fsr2_to_fsr1
	
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; fsr12_to_ustat points FSR1 at UBDESC for current xfer, then FSR2 at its buffer
fsr12_to_ustat
	rlncf	USTAT, W, ACCESS	; Determine which buffer just went INT

	andlw	0xfc				; Mask result to give us offset in UBDESC
	movwf	FSR1L, ACCESS		; Point FSR1 to UBDESC
	movlw	high UBDESC
	movwf	FSR1H, ACCESS
fsr2_to_fsr1
	; point fsr2 at fsr1's buffer
	movlw	2
	movff	PLUSW1, FSR2L		; FSR2 points to DESCADRH:DESCADRL
	movlw	3
	movff	PLUSW1, FSR2H
	return	

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; handover_data0 send WREG bytes to USB FSR1, and marks the buffer with data0
handover_data0:
	movwf	PREINC1, ACCESS
	decf	FSR1L, F, ACCESS
	movlw	0x80 | USBPARITY
	movwf	INDF1, ACCESS
	return
	
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; handover_data1 send WREG bytes to USB FSR1, and marks the buffer with data1
handover_data1:
	movwf	PREINC1, ACCESS
	decf	FSR1L, F, ACCESS
	movlw	0xC0 | USBPARITY
	movwf	INDF1, ACCESS
	return

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; DoBaudDiv - perform 24 bit division to get 16 bit baudrate divider
; Notes: ideally should be SPBRG = (12,000,000/baudrate)-1 rounded to nearest
; integer.  However, to shave instructions, instead I'm doing the following
; (12,001,024/baudrate)-1 with no rounding.  Increasing the numerator nudges
; answers to be slightly higher, achieving proper rounding for standard baud
; rates.
remainH	EQU	tempreg0
remainM	EQU	tempreg1
remainL	EQU	tempreg2
BCount	EQU	tempreg3
denomH	EQU	baudrate2
denomM	EQU	baudrate1
denomL	EQU	baudrate0
resultM	EQU	SPBRGH
resultL	EQU	SPBRG
		
DoBaudDiv:		
	movlw	0xB7			; copy Numerator into Shift Holding ram registers
	movwf	remainL,BANKED
	movlw	0x1F			;0x1B is for actual clock, 1F is for rounding hack
	movwf	resultM,ACCESS
	clrf	resultL,ACCESS

	clrf	remainH,BANKED	; clear final Remainder Ram locations
	clrf	remainM,BANKED

	movlw	.16				; set decimal 16 loop count (only need 16 bit answer)
	movwf	BCount,BANKED

dloop:
	bcf		STATUS, C,ACCESS	; bit clear Carry Flag in STATUS register
	rlcf	resultL, f,ACCESS	; Shift numerator(dividend) Left to move
	rlcf	resultM, f,ACCESS	; next bit to remainder

	rlcf	remainL, f,BANKED	; shift carry (next Dividend bit) into remainder
	rlcf	remainM, f,BANKED
	rlcf	remainH, f,BANKED

	movf	denomH,w,BANKED
	subwf	remainH,w,BANKED	; subtract denominator from Remainder HIGH byte.
	bnz		chklarger			; if result was non-ZERO find out which is larger

	movf	denomM,w,BANKED
	subwf	remainM,w,BANKED	; subtract denominator from Remainder MIDDLE byte.
	bnz		chklarger			; if result was non-ZERO find out which is larger

	movf	denomL,w,BANKED
	subwf	remainL,w,BANKED	; subtract denominator from Remainder LOW byte.

chklarger:
	bnc		pushres			; CARRY clear? then denom < reemainder, skip subtraction
      
	movf	denomL,w,BANKED			 
	subwf	remainL, f,BANKED	; Subtract denominator from remainder value in Low Byte
	movf	denomM,w,BANKED
	subwfb	remainM, f,BANKED	; Subtract denominator from remainder value in Mid Byte
	movf	denomH,w,BANKED
	subwfb	remainH, f,BANKED	; Subtract denominator from remainder value in High Byte

	bsf		resultL, 0,ACCESS	; set LSb in previously shifted result
pushres:
	decfsz	BCount, f,BANKED	; decrement the Loop Bit Counter
	bra		dloop

	movlw	0x00				; decrement result, performing job of rounding
	decf	resultL, f, ACCESS	; this actually decrements past a rounded result
	subwfb	resultM, f, ACCESS	; brg needs to be 1 less then the division result
done:
	return

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

        end
